from tensorflow_similarity.types import FloatTensor, IntTensor
from tqdm.auto import tqdm
import tensorflow_datasets as tfds
from typing import Callable, Optional, Union, List, Sequence

from .samplers import Augmenter
from .memory_samplers import MultiShotMemorySampler


class TFDatasetMultiShotMemorySampler(MultiShotMemorySampler):
    def __init__(self,
                 dataset_name: str,
                 classes_per_batch: int,
                 x_key: str = "image",
                 y_key: str = "label",
                 splits: Union[str, List[str]] = ["train", "test"],
                 examples_per_class_per_batch: int = 2,
                 steps_per_epoch: int = 1000,
                 class_list: Sequence[int] = None,
                 total_examples_per_class: int = None,
                 preprocess_fn: Optional[Callable] = None,
                 augmenter: Optional[Augmenter] = None,
                 warmup: int = -1):
        """Create a Multishot in memory sampler from a dataset downloaded from
        the [TensorFlow datasets catalogue](https://www.tensorflow.org/datasets/catalog/)

        The sampler ensures that each batch is well balanced by ensure that
        each batch aims to contains `example_per_class` examples
        of `classes_per_batch` classes.

        The `batch_size` used during training will be equal to:
        `classes_per_batch * example_per_class` unless an `augmenter` that alters
        the number of examples returned is used. Then the batch_size is a
        function of how many augmented examples are returned by
        the `augmenter`.

        Multishot samplers are to be used when you have multiple
        examples for the same class. If this is not the case, then see
        the [SingleShotMemorySampler()](single_memory.md) for using single
        example with augmentation.

        Memory samplers are good for datasets that fit in memory. If you have
        larger ones that needs to sample from disk then use the
        [TFRecordDatasetSampler()](tfdataset_sampler.md)

        Args:
            dataset_name: the name of the dataset to download and uses as
            referenced in the TensorFlow catalog dataset page.

            x_key: name of the dictonary key that contains the data to feed as
            model input as referenced in the TensorFlow catalog dataset page.
            Defaults to "image".

            y_key: name of the dictonary key that contains the labels as
            referenced in the TensorFlow catalog dataset page.
            Defaults to "label".

            splits: which dataset split(s) to use. Default
            is ["train", "train"] Refersto the catalog page for
            the list of available splits.

            examples_per_class_per_batch: How many example of each class to
            use per batch. Defaults to 2.

            steps_per_epoch: How many steps/batches per epoch.
            Defaults to 1000.

            class_list: Filter the list of examples to only keep those who
            belong to the supplied class list.

            total_examples_per_class: Restrict the number of examples for EACH
            class to total_examples_per_class if set. If not set, all the
            available examples are selected. Defaults to None - no selection.

            preprocess_fn: Preprocess function to apply to the dataset after
            download e.g to resize images. Takes an x and a y.
            Defaults to None.

            augmenter: A function that takes a batch in and return a batch out.
            Can alters the number of examples returned which in turn change the
            batch_size used. Defaults to None.

            warmup: Keep track of warmup epochs and let the augmenter knows
            when the warmup is over by passing along with each batch data a
            boolean `is_warmup`. See `self.get_examples()` Defaults to 0.
        """

        # dealing with users passing a single split e.g "train"
        # instead of ["train"]
        if isinstance(splits, str):
            splits = [splits]

        # we are reusing the memory sampler so "all we need to do" is convert
        # the splits into memory arrays and call the super.
        x = []
        y = []
        for split in splits:
            ds, ds_info = tfds.load(dataset_name, split=split, with_info=True)

            if x_key not in ds_info.features:
                raise ValueError("x_key not found - available features are:",
                                 str(ds_info.features.keys()))
            if y_key not in ds_info.features:
                raise ValueError("y_key not found - available features are:",
                                 str(ds_info.features.keys()))

            pb = tqdm(total=ds_info.splits[split].num_examples,
                      desc="converting %s" % split)

            for e in ds:
                x.append(e[x_key])
                y.append(e[y_key])
                pb.update()
            pb.close()

        # apply preprocess if needed.
        if preprocess_fn:
            x_pre = []
            y_pre = []
            for idx in tqdm(range(len(x)), desc="Preprocessing data"):
                xb, yb = preprocess_fn(x[idx], y[idx])
                x_pre.append(xb)
                y_pre.append(yb)

            x = x_pre
            y = y_pre

        # delegate to the base memorysample
        super().__init__(x,
                         y,
                         classes_per_batch=classes_per_batch,
                         examples_per_class_per_batch=examples_per_class_per_batch,  # noqa
                         steps_per_epoch=steps_per_epoch,
                         class_list=class_list,
                         total_examples_per_class=total_examples_per_class,
                         augmenter=augmenter,
                         warmup=warmup)



from typing import List, Union
from pathlib import Path

import tensorflow as tf
from tensorflow.keras.callbacks import Callback

from tensorflow_similarity.types import Tensor
from tensorflow_similarity.evaluators import MemoryEvaluator
from tensorflow_similarity.metrics import EvalMetric, make_metric
from .types import FloatTensor, IntTensor


class EvalCallback(Callback):
    """Epoch end evaluation callback that build a test index and evaluate
    model performance on it.

    This evaluation only run at epoch_end as it is computationally very
    expensive.

    """

    def __init__(self,
                 queries: Tensor,
                 query_labels: List[int],
                 targets: Tensor,
                 target_labels: List[int],
                 distance: str = 'cosine',
                 metrics: List[Union[str, EvalMetric]] = ['accuracy', 'mean_rank'],  # noqa
                 tb_logdir: str = None,
                 k: int = 1):
        """Evaluate model matching quality against a validation dataset at
        epoch end.

        Args:
            queries: Test examples that will be tested against the built index.

            query_labels: Queries nearest neighboors expected labels.

            targets: Examples that are indexed.

            target_labels: Target examples labels.

            distance: Distance function used to compute pairwise distance
            between examples embeddings.

            metrics: List of [EvalMetrics](eval_metrics.md) to be computed
            during the evaluation. Defaults to ['accuracy', 'mean_rank'].
            embedding to evaluate.

            tb_logdir: Where to write TensorBoard logs. Defaults to None.

            k: How many neigboors to retrive for evaluation. Defaults to 1.

        """
        super().__init__()
        self.queries = queries
        self.queries_labels = query_labels
        self.targets = targets
        self.targets_labels = target_labels
        self.distance = distance
        self.k = k
        self.index_size = len(target_labels)
        self.evaluator = MemoryEvaluator()
        # typing requires this weird formulation of creating a new list
        self.metrics: List[Union[str, EvalMetric]] = [make_metric(m) for m in metrics] # noqa

        if tb_logdir:
            tb_logdir = str(Path(tb_logdir) / 'index/')
            self.tb_writer = tf.summary.create_file_writer(tb_logdir)
            print('TensorBoard logging enable in %s' % tb_logdir)
        else:
            self.tb_writer = None

    def on_epoch_end(self, epoch: int, logs: dict = None):
        if logs is None:
            logs = {}

        # reset the index
        self.model.reset_index()

        # rebuild the index
        self.model.index(self.targets, self.targets_labels, verbose=0)

        lookups = self.model.lookup(self.queries, verbose=0)
        results = self.evaluator.evaluate(self.index_size, self.metrics,
                                          self.queries_labels, lookups)

        mstr = []
        for k, v in results.items():
            logs[k] = v
            mstr.append(f'{k}:{v:0.4f}')
            if self.tb_writer:
                with self.tb_writer.as_default():
                    tf.summary.scalar(k, v, step=epoch)

        print(' - '.join(mstr))


class SplitValidationLoss(Callback):
    """A split validation callback.

    This callback will split the validation data into two sets.

        1) The set of classes seen during training.
        2) The set of classes not seen during training.

    The callback will then compute a separate validation for each split.

    This is useful for separately tracking the validation loss on the seen and
    unseen classes and may provide insight into how well the embedding will
    generalize to new classes.

    Attributes:
        x_known: The set of examples from the known classes.
        y_known: The labels associated with the known examples.
        x_unknown: The set of examples from the unknown classes.
        y_unknown: The labels associated with the unknown examples.
    """
    def __init__(self,
                 x: FloatTensor,
                 y: IntTensor,
                 known_classes: IntTensor):
        """Creates the validation callbacks.

        Args:
            x: Validation data.
            y: Validation labels.
            known_classes: The set of classes seen during training.
        """
        super().__init__()

        # Create separate validation sets for the known and unknown classes
        y = tf.cast(y, dtype=tf.int32)
        known_classes = tf.cast(known_classes, dtype=tf.int32)
        known_classes = tf.squeeze(known_classes)

        # Use broadcasting to do a y X known_classes equality check. By adding
        # a dim to the start of known_classes and a dim to the end of y, this
        # essentially checks `for ck in known_classes: for cy in y: ck == cy`.
        # We then reduce_any to find all rows in y that match at least one
        # class in known_classes.
        # See https://numpy.org/doc/stable/user/basics.broadcasting.html
        broadcast_classes = tf.expand_dims(known_classes, axis=0)
        broadcast_labels = tf.expand_dims(y, axis=-1)
        known_mask = tf.math.reduce_any(
                broadcast_classes == broadcast_labels, axis=1)
        known_idxs = tf.squeeze(tf.where(known_mask))
        unknown_idxs = tf.squeeze(tf.where(~known_mask))

        with tf.device("/cpu:0"):
            known_idxs_nd = tf.reshape(known_idxs, (-1, 1))
            self.x_known = tf.gather_nd(x, indices=known_idxs_nd)
            self.y_known = tf.gather(y, indices=known_idxs)

            unknown_idxs_nd = tf.reshape(unknown_idxs, (-1, 1))
            self.x_unknown = tf.gather_nd(x, indices=unknown_idxs_nd)
            self.y_unknown = tf.gather(y, indices=unknown_idxs)

    def on_epoch_end(self, epoch: int, logs: dict = None):
        _ = epoch
        if logs is None:
            logs = {}
        known_eval = self.model.evaluate(self.x_known, self.y_known, verbose=0)
        unknown_eval = (
                self.model.evaluate(self.x_unknown, self.y_unknown, verbose=0))
        print(f'val_los - known_classes: {known_eval:.4f} - '
              f'unknown_classes: {unknown_eval:.4f}')
        logs['known_val_loss'] = known_eval
        logs['unknown_val_loss'] = unknown_eval



# The file is used to list the public API for document generation purpose.
from tensorflow_similarity import architectures
from tensorflow_similarity import evaluators
from tensorflow_similarity import layers
from tensorflow_similarity import losses
from tensorflow_similarity import metrics
from tensorflow_similarity import matchers
from tensorflow_similarity import models
from tensorflow_similarity import samplers
from tensorflow_similarity import tables
from tensorflow_similarity import visualization
# from tensorflow_similarity import algebra  # not public
from tensorflow_similarity import callbacks
from tensorflow_similarity import distance_metrics
from tensorflow_similarity import distances
from tensorflow_similarity import indexer
from tensorflow_similarity import metrics
# from tensorflow_similarity import types  # not public
from tensorflow_similarity import utils



from abc import ABC, abstractmethod
from tensorflow_similarity.types import FloatTensor
from typing import List, Tuple, Union
from tensorflow_similarity.distances import Distance


class Matcher(ABC):

    @abstractmethod
    def __init__(self,
                 distance: Union[Distance, str],
                 dim: int,
                 verbose: bool,
                 **kwargs):
        """Initializes a nearest neigboors matcher.

        Args:
            distance: the distance used to compute the distance between
            embeddings.

            dim: the size of the embeddings.

            verbose: be verbose.
        """

    @abstractmethod
    def add(self,
            embedding: FloatTensor,
            idx: int,
            verbose: int = 1,
            **kwargs):
        """Add a single embedding to the matcher.

        Args:
            embedding: The embedding to index as computed by
            the similarity model.

            idx: Embedding id as in the index table.
            Returned with the embedding to allow to lookup
            the data associated with a given embedding.

        """

    @abstractmethod
    def batch_add(self,
                  embeddings: FloatTensor,
                  idxs: List[int],
                  verbose: int = 1,
                  **kwargs):
        """Add a batch of embeddings to the matcher.

        Args:
            embeddings: List of embeddings to add to the index.

            idxs (int): Embedding ids as in the index table. Returned with
            the embeddings to allow to lookup the data associated
            with the returned embeddings.

            verbose: Be verbose. Defaults to 1.
        """

    @abstractmethod
    def lookup(self,
               embedding: FloatTensor,
               k: int = 5) -> Tuple[List[int], List[float]]:
        """Find embedding K nearest neighboors embeddings.

        Args:
            embedding: Query embedding as predicted by the model.
            k: Number of nearest neighboors embedding to lookup. Defaults to 5.
        """

    @abstractmethod
    def batch_lookup(self,
                     embeddings: FloatTensor,
                     k: int = 5) -> Tuple[List[List[int]], List[List[float]]]:
        """Find embeddings K nearest neighboors embeddings.

        Args:
            embedding: Batch of query embeddings as predicted by the model.
            k: Number of nearest neighboors embedding to lookup. Defaults to 5.
        """

    @abstractmethod
    def save(self, path: str):
        """Serializes the index data on disk

        Args:
            path: where to store the data
        """

    @abstractmethod
    def load(self, path: str):
        """load index on disk

        Args:
            path: where to store the data
        """



from typing import Dict
import tensorflow as tf
from tensorflow.keras.layers import Layer
from tensorflow.keras import layers
from .types import FloatTensor


@tf.keras.utils.register_keras_serializable(package="Similarity")
class MetricEmbedding(Layer):
    def __init__(self, size: int):
        """Normalized embedding layer
        Args:
            size: size of the embbeding. Usually something like 32, 64 or
            256 floats.
        """
        self.size = size
        self.dense = layers.Dense(size)
        super().__init__()
        # FIXME: enforce the shape
        # self.input_spec = rank2

    def call(self, inputs: FloatTensor) -> FloatTensor:
        x = self.dense(inputs)
        normed_x: FloatTensor = tf.math.l2_normalize(x, axis=1)
        return normed_x

    def get_config(self) -> Dict[str, int]:
        return {'size': self.size}



from os import name
import nmslib
from pathlib import Path
from typing import List, Tuple, Union

from tensorflow_similarity.distances import Distance, distance_canonicalizer
from .matcher import Matcher
from tensorflow_similarity.types import FloatTensor


class NMSLibMatcher(Matcher):
    """
    Efficiently find nearest embeddings by indexing known embeddings and make
    them searchable using the  [Approximate Nearest Neigboors Search](https://en.wikipedia.org/wiki/Nearest_neighbor_search)
    search library [NMSLIB](https://github.com/nmslib/nmslib).
    """

    def __init__(self,
                 distance: Union[Distance, str],
                 dims: int,
                 algorithm: str = 'nmslib_hnsw',
                 **kwargs):

        distance_obj: Distance = distance_canonicalizer(distance)

        # convert to nmslib word
        if distance_obj.name == 'cosine':
            space = 'cosinesimil'
        elif distance_obj.name == 'euclidean':
            space = 'l2'
        elif distance_obj.name == 'manhattan':
            space = 'l1'
        else:
            raise ValueError('Unsupported metric space')

        if algorithm == 'nmslib_hnsw':
            method = 'hnsw'
        else:
            raise ValueError('Unsupported algorithm')

        self._matcher = nmslib.init(method=method, space=space)
        self._matcher.createIndex()


    def add(self,
            embedding: FloatTensor,
            idx: int,
            verbose: int = 1,
            build: bool = True,
            **kwargs):
        """Add an embedding to the index

        Args:
            embedding: The embedding to index as computed by
            the similarity model.

            idx: Embedding id as in the index table.
            Returned with the embedding to allow to lookup
            the data associated with a given embedding.

            verbose: Be verbose. Defaults to 1.

            build: Rebuild the index after the addition.
            Required to make the embedding searchable.
            Set to false to save time between successive addition.
            Defaults to True.

        """
        self._matcher.addDataPoint(idx, embedding)
        if build:
            self._build(verbose=verbose)

    def batch_add(self,
                  embeddings: FloatTensor,
                  idxs: List[int],
                  verbose: int = 1,
                  build: bool = True,
                  **kwargs):
        """Add a batch of embeddings to the matcher.

        Args:
            embeddings: List of embeddings to add to the index.

            idxs (int): Embedding ids as in the index table. Returned with
            the embeddings to allow to lookup the data associated
            with the returned embeddings.

            verbose: Be verbose. Defaults to 1.

            build: Rebuild the index after the addition. Required to
            make the embeddings searchable. Set to false to save
            time between successive addition. Defaults to True.
        """
        # !addDataPoint and addDataPointBAtch have inverted parameters
        if verbose:
            print('|-Adding embeddings to index.')
        self._matcher.addDataPointBatch(embeddings, idxs)

        if build:
            if verbose:
                print('|-Building index.')
            self._build(verbose=verbose)

    def lookup(self,
               embedding: FloatTensor,
               k: int = 5) -> Tuple[List[int], List[float]]:
        """Find embedding K nearest neighboors embeddings.

        Args:
            embedding: Query embedding as predicted by the model.
            k: Number of nearest neighboors embedding to lookup. Defaults to 5.
        """
        idxs: List[int] = []
        distances: List[float] = []
        idxs, distances = self._matcher.knnQuery(embedding, k=k)
        return idxs, distances

    def batch_lookup(self,
                     embeddings: FloatTensor,
                     k: int = 5) -> Tuple[List[List[int]], List[List[float]]]:
        """Find embeddings K nearest neighboors embeddings.

        Args:
            embedding: Batch of query embeddings as predicted by the model.
            k: Number of nearest neighboors embedding to lookup. Defaults to 5.
        """
        batch_idxs = []
        batch_distances = []

        nn = self._matcher.knnQueryBatch(embeddings, k=k)
        for n in nn:
            batch_idxs.append(n[0])
            batch_distances.append(n[1])
        return batch_idxs, batch_distances

    def save(self, path: str):
        """Serializes the index data on disk

        Args:
            path: where to store the data
        """
        fname = self.__make_fname(path)
        self._matcher.saveIndex(fname, save_data=True)

    def load(self, path: str):
        """load index on disk

        Args:
            path: where to store the data
        """
        fname = self.__make_fname(path)
        self._matcher.loadIndex(fname, load_data=True)

    def _build(self, verbose=0):
        """Build the index this is need to take into account the new points
        """
        show = True if verbose else False
        self._matcher.createIndex(print_progress=show)

    def __make_fname(self, path):
        return str(Path(path) / 'index_matcher.bin')



import itertools
import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
from tensorflow_similarity.types import IntTensor


def confusion_matrix(y_pred: IntTensor,
                     y_true: IntTensor,
                     normalize: bool = True,
                     labels: IntTensor = None,
                     title: str = 'Confusion matrix',
                     cmap: str = 'Blues'):
    """Plot confusion matrix

    Args:
        y_pred: Model prediction returned by `model.match()`

        y_true: Expected class_id.

        normalize: Normalizes matrix values between 0 and 1.
        Defaults to True.

        labels: List of class string label to display instead of the class
        numerical ids. Defaults to None.

        title: Title of the confusion matrix. Defaults to 'Confusion matrix'.

        cmap: Color schema as CMAP. Defaults to 'Blues'.
    """

    cm = np.array(tf.math.confusion_matrix(y_true, y_pred))
    accuracy = np.trace(cm) / np.sum(cm).astype('float')
    misclass = 1 - accuracy

    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

    plt.figure(figsize=(8, 6))
    plt.imshow(cm, interpolation='nearest', cmap=cmap)
    plt.title(title)
    plt.colorbar()

    if labels is not None:
        tick_marks = np.arange(len(labels))
        plt.xticks(tick_marks, labels, rotation=45)
        plt.yticks(tick_marks, labels)

    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

    thresh = cm.max() / 1.5 if normalize else cm.max() / 2
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        val = cm[i, j]
        color = "white" if val > thresh else "black"
        txt = "%.2f" % val if val > 0.0 else "0"
        plt.text(j, i, txt, horizontalalignment="center", color=color)

    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel('Predicted label\naccuracy={:0.4f}; misclass={:0.4f}'.format(
        accuracy, misclass))
    plt.show()



from abc import ABC, abstractmethod
from typing import Union, List
import tensorflow as tf
from .types import FloatTensor


class Distance(ABC):
    """
    Note: don't forget to add your distance to the DISTANCES list
    and add alias names in it.

    """
    def __init__(self, name: str, aliases: List[str] = []):
        self.name = name
        self.aliases = aliases

    @abstractmethod
    def call(self, embeddings: FloatTensor) -> FloatTensor:
        """Compute pairwise distances for a given batch.

        Args:
            embeddings: Embeddings to compute the pairwise one.

        Returns:
            FloatTensor: Pairwise distance tensor.
        """

    def __call__(self, embeddings: FloatTensor):
        return self.call(embeddings)

    def __str__(self) -> str:
        return self.name

    def get_config(self):
        return {}


@tf.keras.utils.register_keras_serializable(package="Similarity")
class InnerProductSimilarity(Distance):
    """Compute the pairwise inner product between embeddings.

    The [Inner product](https://en.wikipedia.org/wiki/Inner_product_space) is
    a measure of similarity where the more similar vectors have the largest
    values.

    NOTE! This is not a distance and is likely not what you want to use with
    the built in losses. At the very least this will flip the sign on the
    margin in many of the losses. This is likely meant to be used with custom
    loss functions that expect a similarity instead of a distance.
    """
    def __init__(self):
        "Init Inner product similarity"
        super().__init__('inner_product', ['ip'])

    @tf.function
    def call(self, embeddings: FloatTensor) -> FloatTensor:
        """Compute pairwise similarities for a given batch of embeddings.

        Args:
            embeddings: Embeddings to compute the pairwise one.

        Returns:
            FloatTensor: Pairwise distance tensor.
        """

        tensor = tf.linalg.matmul(embeddings, embeddings, transpose_b=True)
        sims: FloatTensor = tf.reduce_sum(tensor, axis=1, keepdims=True)
        return sims


@tf.keras.utils.register_keras_serializable(package="Similarity")
class CosineDistance(Distance):
    """Compute pairwise cosine distances between embeddings.

    The [Cosine Distance](https://en.wikipedia.org/wiki/Cosine_similarity) is
    an angular distance that varies from 0 (similar) to 1 (dissimilar).
    """
    def __init__(self):
        "Init Cosine distance"
        super().__init__('cosine')

    @tf.function
    def call(self, embeddings: FloatTensor) -> FloatTensor:
        """Compute pairwise distances for a given batch of embeddings.

        Args:
            embeddings: Embeddings to compute the pairwise one.

        Returns:
            FloatTensor: Pairwise distance tensor.
        """
        distances = 1 - tf.linalg.matmul(
                embeddings, embeddings, transpose_b=True)
        min_clip_distances: FloatTensor = tf.math.maximum(distances, 0.0)
        return min_clip_distances


@tf.keras.utils.register_keras_serializable(package="Similarity")
class EuclideanDistance(Distance):
    """Compute pairwise euclidean distances between embeddings.

    The [Euclidean Distance](https://en.wikipedia.org/wiki/Euclidean_distance)
    is the standard distance to measure the line segment between two embeddings
    in the Cartesian point. The larger the distance the more dissimilar
    the embeddings are.

    **Alias**: L2 Norm, Pythagorean
    """
    def __init__(self):
        "Init Euclidean distance"
        super().__init__('euclidean', ['l2', 'pythagorean'])

    @tf.function
    def call(self, embeddings: FloatTensor) -> FloatTensor:
        """Compute pairwise distances for a given batch of embeddings.

        Args:
            embeddings: Embeddings to compute the pairwise one.

        Returns:
            FloatTensor: Pairwise distance tensor.
        """
        squared_norm = tf.math.square(embeddings)
        squared_norm = tf.math.reduce_sum(squared_norm, axis=1, keepdims=True)

        distances: FloatTensor = 2.0 * tf.linalg.matmul(
            embeddings, embeddings, transpose_b=True)
        distances = squared_norm - distances + tf.transpose(squared_norm)

        # Avoid NaN and inf gradients when back propagating through the sqrt.
        # values smaller than 1e-18 produce inf for the gradient, and 0.0
        # produces NaN. All values smaller than 1e-13 should produce a gradient
        # of 1.0.
        dist_mask = tf.math.greater_equal(distances, 1e-18)
        distances = tf.math.maximum(distances, 1e-18)
        distances = tf.math.sqrt(distances) * tf.cast(dist_mask, tf.float32)

        return distances


@tf.keras.utils.register_keras_serializable(package="Similarity")
class SquaredEuclideanDistance(Distance):
    """Compute pairwise squared Euclidean distance.

    The [Sequared Euclidean Distance](https://en.wikipedia.org/wiki/Euclidean_distance#Squared_Euclidean_distance) is
    a distance that varies from 0 (similar) to infinity (dissimilar).
    """
    def __init__(self):
        super().__init__('squared_euclidean', ['sql2', 'sqeuclidean'])

    @tf.function
    def call(self, embeddings: FloatTensor) -> FloatTensor:
        """Compute pairwise distances for a given batch of embeddings.

        Args:
            embeddings: Embeddings to compute the pairwise one.

        Returns:
            FloatTensor: Pairwise distance tensor.
        """
        squared_norm = tf.math.square(embeddings)
        squared_norm = tf.math.reduce_sum(squared_norm, axis=1, keepdims=True)

        distances: FloatTensor = 2.0 * tf.linalg.matmul(
            embeddings, embeddings, transpose_b=True)
        distances = squared_norm - distances + tf.transpose(squared_norm)
        distances = tf.math.maximum(distances, 0.0)

        return distances


@tf.keras.utils.register_keras_serializable(package="Similarity")
class ManhattanDistance(Distance):
    """Compute pairwise Manhattan distances between embeddings.

    The [Manhattan Distance](https://en.wikipedia.org/wiki/Euclidean_distance)
    is the sum of the lengths of the projections of the line segment between
    two embeddings onto the Cartesian axes. The larger the distance the more
    dissimilar the embeddings are.
    """
    def __init__(self):
        "Init Manhattan distance"
        super().__init__('manhattan', ['l1', 'taxicab'])

    @tf.function
    def call(self, embeddings: FloatTensor) -> FloatTensor:
        """Compute pairwise distances for a given batch of embeddings.

        Args:
            embeddings: Embeddings to compute the pairwise one.

        Returns:
            FloatTensor: Pairwise distance tensor.
        """
        x_rs = tf.reshape(embeddings, shape=[tf.shape(embeddings)[0], -1])
        deltas = tf.expand_dims(x_rs, axis=1) - tf.expand_dims(x_rs, axis=0)
        distances: FloatTensor = tf.norm(deltas, 1, axis=2)
        return distances


# List of implemented distances
DISTANCES = [
    InnerProductSimilarity(),
    EuclideanDistance(),
    SquaredEuclideanDistance(),
    ManhattanDistance(),
    CosineDistance()
]


def distance_canonicalizer(user_distance: Union[Distance, str]) -> Distance:
    """Normalize user requested distance to its matching Distance object.

    Args:
        user_distance: Requested distance either by name or by object

    Returns:
        Distance: Requested object name.
    """

    # just return Distance object
    if isinstance(user_distance, Distance):
        # user supplied distance function
        return user_distance

    mapping = {}
    name2fn = {}
    for distance in DISTANCES:
        # self reference
        mapping[distance.name] = distance.name
        name2fn[distance.name] = distance
        # aliasing
        for alias in distance.aliases:
            mapping[alias] = distance.name

    if isinstance(user_distance, str):
        user_distance = user_distance.lower().strip()
        if user_distance in mapping:
            user_distance = mapping[user_distance]
        else:
            raise ValueError('Metric not supported by the framework')

        return name2fn[user_distance]

    raise ValueError('Unknown distance: must either be a MetricDistance\
                     or a known distance function')



from matplotlib import pyplot as plt
from tensorflow_similarity.types import Tensor, Lookup
from typing import List, Dict, Tuple


def viz_neigbors_imgs(example: Tensor,
                      example_class: List[int],
                      neighbors: List[Lookup],
                      class_mapping: Dict = None,
                      fig_size: Tuple = (24, 4),
                      cmap: str = 'viridis'):
    """Display images nearest neighboors

    Args:
        example: The data used as query input.
        example_class: The class of the data used as query
        neighbors: The list of neighbors returned by the lookup()
        class_mapping: Dictionary that map the class numerical ids to a class
        name. If not set, will display the class numerical id.
        Defaults to None.
        fig_size: Size of the figure. Defaults to (24, 4).
        cmap: Default color scheme for black and white images e.g mnist.
        Defaults to 'viridis'.
    """
    num_cols = len(neighbors) + 1
    plt.subplots(1, num_cols, figsize=fig_size)
    plt_idx = 1

    # draw target
    plt.subplot(1, num_cols, plt_idx)
    plt.imshow(example, cmap=cmap)
    plt.xticks([])
    plt.yticks([])

    val = class_mapping[example_class] if class_mapping else str(example_class)
    plt.title(val)
    plt_idx += 1

    for nbg in neighbors:
        plt.subplot(1, num_cols, plt_idx)
        val = class_mapping[nbg.label] if class_mapping else nbg.label
        legend = "%s - d:%.5f" % (val, nbg.distance)
        if nbg.label == example_class:
            color = cmap
        else:
            color = 'Reds'
        plt.imshow(nbg.data, cmap=color)
        plt.title(legend)
        plt.xticks([])
        plt.yticks([])

        plt_idx += 1
    plt.show()



from abc import abstractmethod
import tensorflow as tf
from abc import ABC
from typing import List, Union, Tuple
from .types import Lookup
from .samplers.samplers import Sampler


class EvalMetric(ABC):
    def __init__(self,
                 direction: str,
                 name: str,
                 canonical_name: str,
                 k: int = None,
                 distance_threshold: float = None) -> None:
        if direction not in ['min', 'max']:
            raise ValueError('Unknown direction - must be in {min,max}')
        self.k = k
        self.distance_threshold = distance_threshold
        self.canonical_name = canonical_name
        self.name = self._suffix_name(name, k, distance_threshold)
        self.direction = direction

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return "%s:%s" % (self.canonical_name, self.name)

    def get_config(self):
        return {
            "name": str(self.name),
            "canonical_name": str(self.canonical_name),
            "direction": str(self.direction),
            "k": int(self.k),
            "distance_threshold": float(self.distance_threshold)
        }

    @staticmethod
    def from_config(config):
        metric = make_metric(config['canonical_name'])
        metric.name = config['name']
        metric.k = config['k']
        metric.distance_threshold = config['distance_threshold']
        return metric

    @abstractmethod
    def compute(self,
                max_k: int,
                targets_labels: List[int],
                index_size: int,
                match_ranks: List[int],
                match_distances: List[float],
                lookups: List[List[Lookup]]) -> Union[int, float]:
        """Compute the metric

        Args:
            max_k: Number of neigboors considers during the retrieval.

            targets_labels: Expected labels for the query. One label per query.
            index_size: Total size of the index.

            match_ranks: Minimal rank at which the targeted label is matched.
            For example if the label is matched by the 2nd closest neigboors
            then match rank is 2. If there is no match then value 0.

            match_distances: Minimal distance at which the targeted label is
            matched. Mirror the *match_rank* arg.

            lookups: Full index lookup results to compute more advance metrics.

        Returns:
            metric results.
        """

    def _suffix_name(self,
                     name: str,
                     k: int = None,
                     distance: float = None) -> str:
        "Suffix metric name with k and distance if needed"
        if k and k > 1:
            name = "%s@%d" % (name, k)

        if distance and distance != 0.5:
            name = "%s:%f" % (name, distance)

        return name

    def filter_ranks(self,
                     match_ranks: List[int],
                     match_distances: List[float],
                     min_rank: int = 1,
                     max_rank: int = None,
                     distance: float = None) -> List[int]:
        """Filter match ranks to only keep matches between `min_rank`
        and `max_rank` below a give distance.

        Notes:
            Neigboors are order by ascending distance.


        Args:
            match_ranks: Min rank at which the embedding match the correct
            label. For example, rank 1 is the closest neigboor,
            Rank 5 is the 5th neigboor. There might be neigboors with
            higher ranks that are from a different class.

            min_rank: Minimal rank to keep (inclusive). Defaults to 1.

            max_rank: Max rank to keep (inclusive). Defaults to None.
            If None will keep all ranks above `min_rank`

        Returns:
            filtered ranks as a dense array with missing elements
            removed. len(filtered_ranks) <= len(match_ranks)
        """
        filtered_ranks = []
        if not max_rank:
            max_rank = max(match_ranks)

        if not distance:
            distance = max(match_distances)

        # print('max_rank', max_rank, 'max_distance', distance)
        for idx, r in enumerate(match_ranks):
            elt_distance = match_distances[idx]
            if (min_rank <= r <= max_rank) and (elt_distance <= distance):
                # print(distance, elt_distance)
                filtered_ranks.append(r)
            # else:
            #    print('missed', idx, r, elt_distance)
        if not len(filtered_ranks):
            return [-1]

        return filtered_ranks

    def compute_retrival_metrics(self,
                                 targets_labels: List[int],
                                 lookups: List[List[Lookup]],  # noqa
                                 ) -> Tuple[int, int, int, int]:
        true_positives = 0
        false_positives = 0
        true_negative = 0
        false_negative = 0
        for lidx, lookup in enumerate(lookups):
            for n in lookup[:self.k]:
                if self.distance_threshold and n.distance > self.distance_threshold:  # noqa
                    if n.label == targets_labels[lidx]:
                        false_negative += 1
                    else:
                        true_negative += 1
                else:
                    if n.label == targets_labels[lidx]:
                        true_positives += 1
                    else:
                        false_positives += 1
        return true_positives, false_positives, false_negative, true_negative


class MeanRank(EvalMetric):
    def __init__(self, name='mean_rank') -> None:
        super().__init__(direction='min',
                         canonical_name='mean_rank',
                         name=name)

    def compute(self, max_k: int, targets_labels: List[int], index_size: int,
                match_ranks: List[int],
                match_distances: List[float],
                lookups: List[List[Lookup]]) -> float:

        # remove unmatched elements (rank 0)
        matches = self.filter_ranks(match_ranks,
                                    match_distances,
                                    max_rank=max_k)
        # ! must cast matches before computing to avoid rounding
        return float(tf.reduce_mean(tf.cast(matches, dtype='float')))


class MinRank(EvalMetric):
    def __init__(self, name='min_rank') -> None:
        super().__init__(direction='min', canonical_name='min_rank', name=name)

    def compute(self, max_k: int, targets_labels: List[int],
                index_size: int, match_ranks: List[int],
                match_distances: List[float],
                lookups: List[List[Lookup]]) -> int:

        # remove unmatched elements (rank 0)
        matches = self.filter_ranks(match_ranks,
                                    match_distances,
                                    max_rank=max_k)
        return int(tf.reduce_min(matches))


class MaxRank(EvalMetric):
    def __init__(self, name='max_rank') -> None:
        super().__init__(direction='min', canonical_name='max_rank', name=name)

    def compute(self, max_k: int, targets_labels: List[int],
                index_size: int, match_ranks: List[int],
                match_distances: List[float],
                lookups: List[List[Lookup]]) -> int:

        # remove unmatched elements (rank 0)
        matches = self.filter_ranks(match_ranks,
                                    match_distances,
                                    max_rank=max_k)
        return int(tf.reduce_max(matches))


class Accuracy(EvalMetric):
    """How many correct matches are returned for the given set pf parameters
    Probably the most important metric. Binary accuracy and match() is when
    k=1.

    num_matches / num_queries

    """
    def __init__(self,
                 distance_threshold: float = None,
                 k: int = 1,
                 name='accuracy') -> None:
        super().__init__(direction='max',
                         name=name,
                         canonical_name='accuracy',
                         k=k,
                         distance_threshold=distance_threshold)

    def compute(self, max_k: int, targets_labels: List[int], index_size: int,
                match_ranks: List[int], match_distances: List[float],
                lookups: List[List[Lookup]]) -> float:
        matches = self.filter_ranks(match_ranks,
                                    match_distances,
                                    max_rank=self.k,
                                    distance=self.distance_threshold)

        num_queries = len(targets_labels)
        num_matches = len(matches)
        return num_matches / num_queries


# FIXME:precision@k and precision:T to be written.
# ThresholdPrecision and RankPrecision()
class FIXMEPrecision(EvalMetric):
    """ Compute the precision of the matches.

    Notes: precision computation for similarity is different from
    the classic version and also tricky to get right so please refers
    to the tf.similarity paper for a complete discussion of how it works.

    Precision formula: `true_positives / (true_positives + false_positives)`


    Reference: TBD
    """
    def __init__(self,
                 distance_threshold: float = 0.5,
                 k: int = 1,
                 name='precision') -> None:

        super().__init__(direction='max',
                         name=name,
                         canonical_name='precision',
                         k=k,
                         distance_threshold=distance_threshold)

    def compute(self, max_k: int, targets_labels: List[int], index_size: int,
                match_ranks: List[int], match_distances: List[float],
                lookups: List[List[Lookup]]) -> float:

        tp, fp, _, _ = self.compute_retrival_metrics(targets_labels, lookups)

        denominator = tp + fp
        if denominator:
            return tp / denominator
        else:
            return 1.0


class Recall(EvalMetric):
    """Computing matcher recall at k for a given distance threshold

    Recall formula: num_misses / num_queries
    """
    def __init__(self,
                 distance_threshold: float = 0.5,
                 k: int = 1,
                 name='recall') -> None:
        super().__init__(direction='max',
                         name=name,
                         canonical_name='recall',
                         k=k,
                         distance_threshold=distance_threshold)

    def compute(self, max_k: int, targets_labels: List[int], index_size: int,
                match_ranks: List[int], match_distances: List[float],
                lookups: List[List[Lookup]]) -> float:

        matches = self.filter_ranks(match_ranks,
                                    match_distances,
                                    max_rank=self.k,
                                    distance=self.distance_threshold)
        num_queries = len(targets_labels)
        num_matches = len(matches)
        misses = num_queries - num_matches
        return misses / num_queries


class F1Score(EvalMetric):
    """Compute the F1 score, also known as balanced F-score or F-measure at K
    below a distance threshold.

    https://scikit-learn.org/stable/modules/generated/sklearn.metrics.f1_score.html

    notes:
        **This is not the traditional formula** its over query set
        not index set. To be meaningful this assumes that each of the
        query have at least 1 matching example in the index.
        See paper for details

    f1_score formula is: `2 * (accuracy * recall) / (accuracy + recall)`

    """
    def __init__(self,
                 distance_threshold: float = 0.5,
                 k: int = 1,
                 name='f1_score') -> None:
        super().__init__(direction='max',
                         name=name,
                         canonical_name='f1_score',
                         k=k,
                         distance_threshold=distance_threshold)

    def compute(self, max_k: int, targets_labels: List[int], index_size: int,
                match_ranks: List[int], match_distances: List[float],
                lookups: List[List[Lookup]]) -> float:

        matches = self.filter_ranks(match_ranks,
                                    match_distances,
                                    max_rank=self.k,
                                    distance=self.distance_threshold)

        num_queries = len(targets_labels)
        num_matches = len(matches)
        num_misses = num_queries - num_matches

        precision = num_matches / num_queries
        recall = num_misses / num_queries

        return 2 * (precision * recall) / (precision + recall)


def make_metric(metric: Union[str, EvalMetric]) -> EvalMetric:
    """Covert metric from str name to object if needed.

    Args:
        metric (Union[str, EvalMetric]): EvalMetric() or metric name.

    Raises:
        ValueError: metric name is invalid.

    Returns:
        EvalMetric: Instanciated metric if needed.
    """
    # ! Metrics must be non-instanciated.
    METRICS_ALIASES = {
        # accuracy
        "accuracy": Accuracy(),
        "acc": Accuracy(),
        # recall
        "recall": Recall(),
        # f1 score
        "f1_score": F1Score(),
        "f1": F1Score(),
        # mean rank
        "mean_rank": MeanRank(),
        "meanrank": MeanRank(),
        # max rank
        "max_rank": MaxRank(),
        "maxrank": MaxRank(),
        # min rank
        "min_rank": MinRank(),
        "minrank": MinRank(),
    }

    if isinstance(metric, EvalMetric):
        return metric
    elif isinstance(metric, str):
        if metric.lower() in METRICS_ALIASES:
            return METRICS_ALIASES[metric.lower()]
        else:
            raise ValueError('Unknown metric name:', metric, ' typo?')
    else:
        raise ValueError('metrics must be a str or a Evalmetric Object')


def make_metrics(metrics: List[Union[str, EvalMetric]]) -> List[EvalMetric]:
    """Convert a list of mixed metrics name and object to a list
    of EvalMetrics

    Args:
        metrics (List[Union[str, EvalMetric]]): Metric and Metric names

    Returns:
        List[EvalMetric]: Metrics objects
    """
    eval_metrics = []
    for m in metrics:
        eval_metrics.append(make_metric(m))
    return eval_metrics


def batch_class_ratio(sampler: Sampler, num_batches: int = 100) -> float:
    """Computes the average number of examples per class within each batch.

    Similarity learning requires at least 2 examples per class in each batch.
    This is needed in order to construct the triplets. This function
    provides the average number of examples per class within each batch and
    can be used to check that a sampler is working correctly.

    The ratio should be >= 2.

    Args:
        sampler: A tf.similarity sampler object.
        num_batches: The number of batches to sample.

    Returns:
        The average number of examples per class.
    """
    ratio = 0
    for batch_count, (_, y) in enumerate(sampler):
        if batch_count < num_batches:
            batch_size = tf.shape(y)[0]
            num_classes = tf.shape(tf.unique(y)[0])[0]
            ratio += tf.math.divide(batch_size, num_classes)
        else:
            break

    return float(ratio/(batch_count+1))



from typing import Tuple
import tensorflow as tf
from tensorflow_similarity.metrics import batch_class_ratio
from tensorflow_similarity.metrics import MaxRank
from tensorflow_similarity.metrics import MeanRank
from tensorflow_similarity.metrics import MinRank
from tensorflow_similarity.samplers.samplers import Sampler
from tensorflow_similarity.types import Lookup, Tensor


MAX_K = 0
TARGETS_LABELS = 1
NUM_MATCHED = 2
NUM_UNMATCHED = 3
INDEX_SIZE = 4
MATCH_RANKS = 5
MATCH_DISTANCES = 6
MATCH_LABELS = 7

TEST_VECTORS = [[
    3,  # max_k
    [1, 2, 3, 4],  # targets_labels
    30,  # index size
    [2, 1, 2, 0],  # match_ranks
    [0.1, 0.2, 0.3, 0],  # match_distances
    [  # lookups
        [
            Lookup(rank=0, label=21, distance=0.01),
            Lookup(rank=1, label=1, distance=0.1)
         ],
        [
            Lookup(rank=0, label=2, distance=0.2),
            Lookup(rank=1, label=22, distance=0.22)
         ],
        [
            Lookup(rank=0, label=23, distance=0.01),
            Lookup(rank=1, label=3, distance=0.3)
         ],
        [
            Lookup(rank=0, label=24, distance=0.44),
            Lookup(rank=1, label=44, distance=0.54)
         ]
    ]
]]


def compute_vector(m, v):
    return m.compute(v[0], v[1], v[2], v[3], v[4], v[5])


def test_mean_rank():
    expected = [1.67]
    m = MeanRank()

    # check name
    assert m.name == 'mean_rank'

    # check computation
    for idx, v in enumerate(TEST_VECTORS):
        res = compute_vector(m, v)
        assert round(res, 2) == expected[idx]


def test_min_rank():
    expected = [1]
    m = MinRank()

    # check name
    assert m.name == 'min_rank'

    # check computation
    for idx, v in enumerate(TEST_VECTORS):
        res = compute_vector(m, v)
        assert res == expected[idx]


def test_max_rank():
    expected = [2]
    m = MaxRank()

    # check name
    assert m.name == 'max_rank'

    # check computation
    for idx, v in enumerate(TEST_VECTORS):
        res = compute_vector(m, v)
        assert res == expected[idx]


def test_f1_score():
    # expected = [3]
    # m = F1Score()

    # # check name
    # assert m.name == 'f1_score'

    # # check computation
    # for idx, v in enumerate(TEST_VECTORS):
    #     res = compute_vector(m, v)
    #     # FIXME compute the real value and check for it
    #     # assert res == expected[idx]
    pass


# lower level functions
def test_filter_rank_k():
    match_ranks = [1, 2, 3, 1, 2, 3]
    match_distances = [0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
    m = MinRank()
    for k in range(1, 3):
        matches = m.filter_ranks(match_ranks,
                                 match_distances,
                                 max_rank=k,
                                 distance=1)
        assert min(matches) == 1
        assert max(matches) == k


def test_filter_rank_distance():
    match_ranks = [1, 2, 3, 1, 2, 3]
    match_distances = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]
    m = MinRank()
    for i in range(1, 7):
        distance = i / 10
        matches = m.filter_ranks(match_ranks,
                                 match_distances,
                                 max_rank=10,
                                 distance=distance)
        assert len(matches) == i


def test_batch_class_ratio():
    class MockSampler(Sampler):

        def __init__(self):
            pass

        def __len__(self):
            return 1

        def __getitem__(self, *args) -> Tuple[Tensor, Tensor]:
            return (tf.ones(8),  tf.constant([0, 0, 1, 1, 2, 2, 3, 3]))

        def get_examples(self):
            pass

    result = batch_class_ratio(MockSampler())
    assert result == 2.0



from tensorflow_similarity.samplers import TFDatasetMultiShotMemorySampler
import pytest


def test_basic():
    dataset_name = "mnist"
    sampler = TFDatasetMultiShotMemorySampler(dataset_name=dataset_name,
                                              classes_per_batch=10)
    batch = sampler.generate_batch(42)
    assert batch[0].shape == (20, 28, 28, 1)

def test_wrong_key():
    dataset_name = "mnist"

    # X
    with pytest.raises(ValueError):
        TFDatasetMultiShotMemorySampler(dataset_name=dataset_name,
                                        classes_per_batch=4,
                                        x_key='error')
    # Y
    with pytest.raises(ValueError):
        TFDatasetMultiShotMemorySampler(dataset_name=dataset_name,
                                        classes_per_batch=4,
                                        y_key='error')



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""PN Loss foudn in Ivis
    Szubert, B., Cole, J.E., Monaco, C. et al.
    Structure-preserving visualisation of high dimensional single-cell dataset
    Sci Rep 9, 8914 (2019). https://doi.org/10.1038/s41598-019-45301-0
"""

import tensorflow as tf
from typing import Any, Callable, Union

from tensorflow_similarity.distances import Distance, distance_canonicalizer
from tensorflow_similarity.algebra import build_masks
from tensorflow_similarity.types import FloatTensor, IntTensor
from .utils import negative_distances, positive_distances, compute_loss
from .metric_loss import MetricLoss


@tf.keras.utils.register_keras_serializable(package="Similarity")
@tf.function
def pn_loss(labels: IntTensor,
            embeddings: FloatTensor,
            distance: Callable,
            positive_mining_strategy: str = 'hard',
            negative_mining_strategy: str = 'semi-hard',
            soft_margin: bool = False,
            margin: float = 1.0) -> Any:
    """Positive Negative loss computations

    Based on the pn loss used in IVIS.


    Args:
        labels: labels associated with the embed

        embeddings: Embedded examples.

        distance: Which distance function to use to compute the pairwise
        distances between embeddings. Defaults to 'cosine'.

        positive_mining_strategy: What mining strategy to use to select
        embedding from the same class. Defaults to 'hard'.
        Available: {'easy', 'hard'}

        negative_mining_strategy: What mining strategy to use for select the
        embedding from the different class. Defaults to 'semi-hard'.
        Available: {'hard', 'semi-hard', 'easy'}

        soft_margin: [description]. Defaults to True. Use a soft margin
        instead of an explicit one.

        margin: Use an explicit value for the margin term. Defaults to 1.0.

    Returns:
        Loss: The loss value for the current batch.
    """

    # [Label]
    # ! Weirdness to be investigated
    # do not remove this code. It is actually needed for specific situation
    # Reshape label tensor to [batch_size, 1] if not already in that format.
    # labels = tf.reshape(labels, (labels.shape[0], 1))
    batch_size = tf.size(labels)

    # [distances]
    pairwise_distances = distance(embeddings)

    # [masks]
    positive_mask, negative_mask = build_masks(labels, batch_size)

    # [Positive distance computation]
    pos_distances, pos_idxs = positive_distances(
            positive_mining_strategy,
            pairwise_distances,
            positive_mask,
    )

    # [Negative distances computation]
    neg_distances, neg_idxs = negative_distances(
            negative_mining_strategy,
            pairwise_distances,
            negative_mask,
            positive_mask,
            batch_size,
    )

    # Compute the distance between the pairs of positive and negative examples.
    # Then take the min(pos_neg_dist, anchor_neg_dist) as the neg_distances.
    # This encourages both the anchor and the positives to be far from the
    # negative.
    pn_pairs = tf.stack([pos_idxs, neg_idxs], axis=1)
    pn_distances = tf.gather_nd(pairwise_distances, pn_pairs)
    pn_distances = tf.reshape(pn_distances, [-1, 1])
    neg_distances = tf.math.minimum(pn_distances, neg_distances)

    # [PN loss computation]
    pn_loss = compute_loss(pos_distances, neg_distances, soft_margin, margin)

    pn_loss = tf.reduce_mean(pn_loss)
    return pn_loss


@tf.keras.utils.register_keras_serializable(package="Similarity")
class PNLoss(MetricLoss):
    """Computes the PN loss in an online fashion.

    This loss encourages the positive distances between a pair of embeddings
    with the same labels to be smaller than the minimum negative distances
    between pair of embeddings of different labels. Additionally, both the
    anchor and the positive embeddings are encouraged to be far from the
    negative embeddings. This is accomplished by taking the
    min(pos_neg_dist, anchor_neg_dist) and using that as the negative distance
    in the triplet loss.

    See PN Loss Ivis:
    Szubert, B., Cole, J.E., Monaco, C. et al.
    Structure-preserving visualisation of high dimensional single-cell dataset
    Sci Rep 9, 8914 (2019). https://doi.org/10.1038/s41598-019-45301-0

    `y_true` must be  a 1-D integer `Tensor` of shape (batch_size,).
    It's values represent the classes associated with the examples as
    **integer  values**.

    `y_pred` must be 2-D float `Tensor`  of L2 normalized embedding vectors.
    you can use the layer `tensorflow_similarity.layers.L2Embedding()` as the
    last layer of your model to ensure your model output is properly
    normalized.
    """

    def __init__(self,
                 distance: Union[Distance, str] = 'cosine',
                 positive_mining_strategy: str = 'hard',
                 negative_mining_strategy: str = 'semi-hard',
                 soft_margin: bool = False,
                 margin: float = 1.0,
                 name: str = 'PNLoss'):
        """Initializes the PN Loss

        Args:
            distance: Which distance function to use to compute
            the pairwise distances between embeddings. Defaults to 'cosine'.

            positive_mining_strategy: What mining strategy to
            use to select embedding from the same class. Defaults to 'hard'.
            available: {'easy', 'hard'}

            negative_mining_strategy: What mining strategy to
            use for select the embedding from the different class.
            Defaults to 'semi-hard'. Available: {'hard', 'semi-hard', 'easy'}

            soft_margin: [description]. Defaults to True.
            Use a soft margin instead of an explicit one.

            margin: Use an explicit value for the margin
            term. Defaults to 1.0.

            name: Loss name. Defaults to PNLoss.

        Raises:
            ValueError: Invalid positive mining strategy.
            ValueError: Invalid negative mining strategy.
            ValueError: Margin value is not used when soft_margin is set
                        to True.
        """

        # distance canonicalization
        distance = distance_canonicalizer(distance)
        self.distance = distance
        # sanity checks

        if positive_mining_strategy not in ['easy', 'hard']:
            raise ValueError('Invalid positive mining strategy.')

        if negative_mining_strategy not in ['easy', 'hard', 'semi-hard']:
            raise ValueError('Invalid negative mining strategy.')

        # Ensure users knows its one or the other
        if margin != 1.0 and soft_margin:
            raise ValueError('Margin value is not used when soft_margin is\
                              set to True')

        super().__init__(pn_loss,
                         name=name,
                         reduction=tf.keras.losses.Reduction.NONE,
                         distance=distance,
                         positive_mining_strategy=positive_mining_strategy,
                         negative_mining_strategy=negative_mining_strategy,
                         soft_margin=soft_margin,
                         margin=margin)



"""
Contrastive learning specialized losses.
"""
from .pn_loss import PNLoss  # noqa
from .triplet_loss import TripletLoss  # noqa
from .metric_loss import MetricLoss  # noqa
from .circle_loss import CircleLoss  # noqa
from .multisim_loss import MultiSimilarityLoss  # noqa



import abc
from typing import Any, Callable, Optional, Tuple

from tensorflow.keras.utils import Sequence
from tensorflow_similarity.types import Tensor

# An Augmenter is a Map TensorLike -> TensorLike. The function must
# implement the following signature:
#
# Args:
#   x: Feature Tensors
#   y: Label Tensors
#   examples_per_class: The number of examples per class.
#   is_warmup: If True, the sampler is still in warmup phase.
# Returns:
#   A Tuple containing the transformed x and y tensors.
Augmenter = (Callable[[Tensor, Tensor, int, bool], Tuple[Tensor, Tensor]])

# Not currently used. Might be useful to allows gradual call of augmenter
Scheduler = Callable[[Any], Any]


class Sampler(Sequence, metaclass=abc.ABCMeta):
    def __init__(
            self,
            classes_per_batch: int,
            examples_per_class_per_batch: int = 2,
            steps_per_epoch: int = 1000,
            augmenter: Optional[Augmenter] = None,
            # scheduler: Optional[Scheduler] = None,
            warmup: int = 0) -> None:
        """Create a dataset sampler that ensure that each batch contains at
        least `example_per_class` examples of `class_per_batch`
        classes. Sampling is needed as contrastive loss requires at
        least two examples of the same class present in a batch.
        The batch_size used during training will be equal to:
        `class_per_batch * example_per_class` unless an `augmenter` that alters
        the number of examples returned is used. Then the batch_size is a
        function of how many augmented examples are returned by
        the `augmenter`.


        Args:
            class_per_batch: Numbers of class to include in a single batch
            example_per_class: How many example to include for each class per
            batch.

            example_per_class: how many example of each class to use per batch.
            Defaults to 2.

            steps_per_epoch: How many steps/batches per epoch. Defaults to
            1000.

            augmenter: A function that takes a batch in and return a batch out.
            Can alters the number of examples returned which in turn change the
            batch_size used. Defaults to None.

            warmup: Keep track of warmup epochs and let the augmenter knows
            when the warmup is over by passing along with each batch data a
            boolean `is_warmup`. See `self.get_examples()` Defaults to 0.
        """

        self.epoch = 0  # track epoch count
        self.classes_per_batch = classes_per_batch
        self.examples_per_class_per_batch = examples_per_class_per_batch
        self.batch_size = classes_per_batch * examples_per_class_per_batch
        self.steps_per_epoch = steps_per_epoch
        self.augmenter = augmenter
        self.warmup = warmup
        self.is_warmup = True if warmup else False

        # Tell the users what to expect as they might be unsure what the batch
        # size will be
        print("\nBatch size is %d (%d class X %d example per class "
              "pre-augmentation" % (self.batch_size, self.classes_per_batch,
                                    self.examples_per_class_per_batch))

    @abc.abstractmethod
    def get_examples(
        self, batch_id: int, num_classes: int, example_per_class: int
    ) -> Tuple[Any, Any]:  # FIXME: Tensor type cause errors.
        """Get the set of examples that would be used to create a single batch.

        Notes:
         - before passing the batch data to TF, the sampler will call the
           augmenter function (if any) on the returned example.

         - A batch_size = num_classes * example_per_class

         - This function must be defined in the subclass.

        Args:
            batch_id: id of the batch in the epoch.
            num_classes: How many class should be present in the examples.
            example_per_class: How many example per class should be returned.

        Returns:
            x, y: batch of examples made of `num_classes` * `example_per_class`
        """
        raise NotImplementedError('must be implemented by subclass')

    # [Shared mechanics]
    def __len__(self) -> int:
        "Return the number of batch per epoch"
        return self.steps_per_epoch

    def on_epoch_end(self) -> None:
        "Keep track of warmup epochs"

        # if self.scheduler:
        #     # fixme scheduler idea
        #     self.scheduler(self.epoch)

        self.epoch += 1
        if self.is_warmup and (self.epoch >= self.warmup):
            print("Warmup complete")
            self.is_warmup = False

    def __getitem__(self, batch_id: int) -> Tuple[Tensor, Tensor]:
        return self.generate_batch(batch_id)

    def generate_batch(self, batch_id: int) -> Tuple[Tensor, Tensor]:
        """Generate a batch of data.


        Args:
            batch_id ([type]): [description]

        Returns:
            x, y: batch
        """

        x, y = self.get_examples(batch_id, self.classes_per_batch,
                                 self.examples_per_class_per_batch)

        # strip examples if needed. This might happen due to rounding
        if len(x) != self.batch_size:
            x = x[:self.batch_size]
            y = y[:self.batch_size]

        if self.augmenter:
            x, y = self.augmenter(x, y, self.examples_per_class_per_batch,
                                  self.is_warmup)
        return x, y
